import logging

import pandas

from migration.connector.destination.base import Destination
from migration.scheduler.task.base_task import Task
from migration.base.status import Status

STATUS_SCHEMA = "cz_migration"
logger = logging.getLogger(__name__)
PK_TABLE_DML_HINT = {'hints': {'cz.sql.allow.insert.table.with.pk': 'true'}}
INCREMENTAL_INDEX = 0


def init_status_table(destination: Destination, project_name: str):
    if destination.name.lower() == "clickzetta":
        destination.execute_sql(f"CREATE SCHEMA IF NOT EXISTS {STATUS_SCHEMA}")

        tables_in_schema = destination.execute_sql(
            f"show tables in  {STATUS_SCHEMA} where table_name like '{project_name}_%'")

        exist_status_tables = [x[1] for x in tables_in_schema]
        exist_status_tables_index = []
        for table in exist_status_tables:
            try:
                exist_status_tables_index.append(int(table.split("_")[-1]))
            except ValueError:
                raise ValueError(f"Table name {table} is not valid")
        exist_status_tables_index.sort()
        table_index = 0 if not exist_status_tables else exist_status_tables_index[-1] + 1
        table_name = f"{STATUS_SCHEMA}.{project_name}_{table_index}"

        ddl = f"""
        CREATE TABLE IF NOT EXISTS {table_name} (
            id INT NOT NULL,
            task_id STRING NOT NULL,
            task_name STRING NOT NULL,
            prject_id STRING NOT NULL,
            task_status STRING NOT NULL,
            task_type STRING NOT NULL,
            task_start_time TIMESTAMP NOT NULL,
            task_end_time TIMESTAMP,
            PRIMARY KEY (id)
        )
        """
        destination.execute_sql(ddl)
        return f"{project_name}_{table_index}"
    elif destination.name.lower() == "doris":
        destination.execute_sql(f"CREATE DATABASE IF NOT EXISTS {STATUS_SCHEMA}")
        tables_in_schema = destination.execute_sql(
            f"show tables in  {STATUS_SCHEMA} where table_name like '{project_name}_%'")
        exist_status_tables = [x['Tables_in_cz_migration'] for x in tables_in_schema]
        exist_status_tables_index = []
        for table in exist_status_tables:
            try:
                exist_status_tables_index.append(int(table.split("_")[-1]))
            except ValueError:
                raise ValueError(f"Table name {table} is not valid")
        exist_status_tables_index.sort()
        table_index = 0 if not exist_status_tables else exist_status_tables_index[-1] + 1
        table_name = f"{STATUS_SCHEMA}.{project_name}_{table_index}"
        ddl = f"""
        CREATE TABLE IF NOT EXISTS {table_name} (
            id INT NOT NULL,
            task_id STRING NOT NULL,
            task_name STRING NOT NULL,
            prject_id STRING NOT NULL,
            task_status STRING NOT NULL,
            task_type STRING NOT NULL,
            task_start_time DATETIME NOT NULL,
            task_end_time DATETIME
        )
        UNIQUE KEY (id)
        DISTRIBUTED BY HASH(id) BUCKETS 6
        PROPERTIES (
            "replication_num" = "1"
        );
        """
        destination.execute_sql(ddl)
        return f"{project_name}_{table_index}"


def update_task_status(destination: Destination, task: Task):
    if destination.name.lower() == "clickzetta":
        destination.execute_sql(
            f"UPDATE {STATUS_SCHEMA}.{task.project_id} SET task_status = '{task.status.value}', task_end_time = cast('{task.end_time}' as timestamp) WHERE id = '{task.status_id}'",
            PK_TABLE_DML_HINT)
    elif destination.name.lower() == "doris":
        destination.execute_sql(
            f"UPDATE {STATUS_SCHEMA}.{task.project_id} SET task_status = '{task.status.value}', task_end_time = cast('{task.end_time}' as datetime) WHERE id = '{task.status_id}'")
    logger.info(f"Updated task {task.id} status to {task.status.value}")


def init_task_status(destination: Destination, task: Task):
    sql = None
    global INCREMENTAL_INDEX
    if destination.name.lower() == "clickzetta":
        sql = f"""
        INSERT INTO {STATUS_SCHEMA}.{task.project_id} (id, task_id, task_name, prject_id, task_status, task_type, task_start_time, task_end_time) 
        values ({INCREMENTAL_INDEX},'{task.id}', '{task.name}','{task.project_id}','{task.status.value}', '{task.task_type.value}', cast('{task.start_time}' as timestamp), null)
        """
        destination.execute_sql(sql, PK_TABLE_DML_HINT)
    elif destination.name.lower() == "doris":
        sql = f"""
        INSERT INTO {STATUS_SCHEMA}.{task.project_id} (id, task_id, task_name, prject_id, task_status, task_type, task_start_time, task_end_time) 
        values ('{INCREMENTAL_INDEX}','{task.id}', '{task.name}','{task.project_id}','{task.status.value}', '{task.task_type.value}', cast('{task.start_time}' as datetime), null)
        """
        destination.execute_sql(sql)
    task.status_id = INCREMENTAL_INDEX
    INCREMENTAL_INDEX += 1
    logger.info(f"Inited task {task.id} status")


def get_last_migration_report(dest: Destination, project_name: str):
    tables_in_schema = dest.execute_sql(
        f"show tables in  {STATUS_SCHEMA} where table_name like '{project_name}_%'")
    if dest.name.lower() == "clickzetta":
        exist_status_tables = [x[1] for x in tables_in_schema]
    elif dest.name.lower() == "doris":
        exist_status_tables = [x['Tables_in_cz_migration'] for x in tables_in_schema]
    else:
        exist_status_tables = [x[0] for x in tables_in_schema]
    exist_status_tables_index = []
    for table in exist_status_tables:
        try:
            exist_status_tables_index.append(int(table.split("_")[-1]))
        except ValueError:
            raise ValueError(f"Table name {table} is not valid")
    exist_status_tables_index.sort()
    result = dest.execute_sql(f"select * from {STATUS_SCHEMA}.{project_name}_{exist_status_tables_index[-1]}")
    import openpyxl
    wb = openpyxl.Workbook(iso_dates=True)
    ws = wb.active
    ws.append(
        ["id", "task_id", "task_name", "prject_id", "task_status", "task_type", "task_start_time", "task_end_time"])
    for row in result:
        if dest.name.lower() == "doirs":
            row = list(row.values())
        if isinstance(row[6], pandas._libs.tslibs.timestamps.Timestamp):
            row_list = list(row)
            row_list[6] = row_list[6].strftime("%Y-%m-%d %H:%M:%S")
            row_list[7] = row_list[7].strftime("%Y-%m-%d %H:%M:%S")
            ws.append(row_list)
        else:
            ws.append(row)
    wb.save(f"{project_name}_migration_report.xlsx")
    logger.info(f"Migration report saved to {project_name}_migration_report.xlsx")


def get_last_migration_status(dest: Destination, project_name: str):
    tables_in_schema = dest.execute_sql(
        f"show tables in  {STATUS_SCHEMA} where table_name like '{project_name}_%'")
    if dest.name.lower() == "clickzetta":
        exist_status_tables = [x[1] for x in tables_in_schema]
    elif dest.name.lower() == "doris":
        exist_status_tables = [x['Tables_in_cz_migration'] for x in tables_in_schema]
    else:
        exist_status_tables = [x[0] for x in tables_in_schema]
    exist_status_tables_index = []
    for table in exist_status_tables:
        try:
            exist_status_tables_index.append(int(table.split("_")[-1]))
        except ValueError:
            raise ValueError(f"Table name {table} is not valid")
    exist_status_tables_index.sort()
    result = dest.execute_sql(f"select * from {STATUS_SCHEMA}.{project_name}_{exist_status_tables_index[-1]}")
    pandas_dict = {'id': [], 'task_id': [], 'task_name': [], 'prject_id': [], 'task_status': [], 'task_type': [],
                   'task_start_time': [], 'task_end_time': []}
    for row in result:
        if dest.name.lower() == "doirs":
            row = list(row.values())
        pandas_dict['id'].append(row[0])
        pandas_dict['task_id'].append(row[1])
        pandas_dict['task_name'].append(row[2])
        pandas_dict['prject_id'].append(row[3])
        pandas_dict['task_status'].append(row[4])
        pandas_dict['task_type'].append(row[5])
        pandas_dict['task_start_time'].append(row[6])
        pandas_dict['task_end_time'].append(row[7])
    import pandas
    dataframe = pandas.DataFrame(pandas_dict)
    print(f"Migration status for project {project_name}:\n")
    from tabulate import tabulate
    print(tabulate(dataframe, headers='keys', tablefmt='pretty'))



import logging
import re
import time
from datetime import datetime

import pymysql
import sqlparse
from pymysqlpool.pool import Pool

from migration.connector.source.base import Source
from migration.base.exceptions import SourceExecutionError
from migration.connector.source.enum import Column, ClusterInfo
import migration.util.object_storage_util as object_storage_util

logger = logging.getLogger(__name__)


class DorisSource(Source):
    def __init__(self, config: dict, meta_conf_path=None, storage_conf_path=None):
        super().__init__('Doris', config)
        self.connection_params = self.get_connection_params()
        self.meta_conf_path = meta_conf_path
        self.storage_config_path = storage_conf_path
        self.pool = Pool(host=self.connection_params['host'], port=self.connection_params['port'],
                         user=self.connection_params['user'], password=self.connection_params['password'],
                         max_size=50, autocommit=True)
        self.pool.init()

    """
    Doris connection parameters is a dict including following keys:
    1. host: host of Doris
    2. port: port for mysql client, default is 9030
    3. database: database name
    4. user: user name
    5. passwd: password
    """

    def get_connection_params(self):
        assert self.config['fe_servers']
        assert self.config['user']
        assert self.config['password']

        return {
            'host': self.config['fe_servers'][0].split(':')[0],
            'port': int(self.config['fe_servers'][0].split(':')[1]),
            'user': self.config['user'],
            'password': self.config['password']
        }

    def connect(self):
        if self.connection is None:
            self.connection = pymysql.connect(**self.connection_params)
            logger.info(f"Connect to Doris {self.connection_params['host']} successfully")

    def get_database_names(self):
        result = self.execute_sql("show databases")
        return [row['Database'] for row in result]

    def get_table_names(self, database_name):
        result = self.execute_sql(f"show tables from {database_name}")
        return [row['Tables_in_data_integration'] for row in result]

    def get_ddl_sql(self, database_name, table_name):
        return self.execute_sql(f"show create table {database_name}.{table_name}")[0]['Create Table']

    def get_table_cluster_info(self, database_name, table_name):
        ddl_sql = self.get_ddl_sql(database_name, table_name)
        cluster_columns = []
        ddl_format_sql = sqlparse.format(ddl_sql, reindent=True, keyword_case='upper')
        match_result = re.match(r'(.*)DISTRIBUTED(.*)BY(.*?)\((.*?)\)(.*)BUCKETS(.*?)(\d+).*', ddl_format_sql, re.S)
        if match_result:
            for cluster_column in match_result.group(4).strip().split(','):
                cluster_columns.append(cluster_column.replace('`', '').strip())
            return ClusterInfo(int(match_result.group(7).strip()), cluster_columns)

        return None

    def get_table_partition_columns(self, database_name, table_name):
        ddl_sql = self.get_ddl_sql(database_name, table_name)
        partition_columns = []
        ddl_format_sql = sqlparse.format(ddl_sql, reindent=True, keyword_case='upper')
        match_result = re.match(r'(.*?)PARTITION(.*?)BY(.*?)\((.*?)\).*', ddl_format_sql, re.S)
        if match_result:
            for partition_column in match_result.group(4).strip().split(','):
                partition_columns.append(partition_column.replace('`', '').strip())
            return partition_columns
        return None

    def get_primary_key(self, database_name, table_name):
        ddl_sql = self.get_ddl_sql(database_name, table_name)
        primary_keys = []
        ddl_format_sql = sqlparse.format(ddl_sql, reindent=True, keyword_case='upper')
        match_result = re.match(r'(.*?)UNIQUE(.*?)KEY(.*?)\((.*?)\).*', ddl_format_sql, re.S)
        if match_result:
            for primary_key in match_result.group(4).strip().split(','):
                primary_keys.append(primary_key.replace('`', '').strip())
            return primary_keys
        return None

    def get_table_columns(self, database_name, table_name) -> list[Column]:
        result = self.execute_sql(f"desc {database_name}.{table_name}")
        table_columns = []
        for row in result:
            table_columns.append(Column(name=row['Field'], type=row['Type'].upper(),
                                        is_null=True if row['Null'].strip() == 'Yes' else False,
                                        default_value=row['Default']))
        return table_columns

    def execute_sql(self, sql, bind_params=None):
        try:
            connection = self.pool.get_conn()
            with connection.cursor() as cur:
                cur.execute(sql, bind_params)
                result = cur.fetchall()
                self.pool.release(connection)
                return result

        except SourceExecutionError as e:
            logger.error(f"Doris connector execute sql {sql} failed, error: {e}")
            raise f"Doris connector execute sql {sql} failed, error: {e}"

    def type_mapping(self):
        return {
            'DATETIME': 'TIMESTAMP',
        }

    def close(self):
        if self.connection is not None:
            self.connection.close()
            self.connection = None
        if self.pool is not None:
            self.pool.destroy()
        logger.info(f"Close connection to Doris {self.connection_params['host']} successfully")

    def unload_data(self, task):
        logger.info(f"Begin to unload data from Doris table {task.name} to object storage")
        file_path = f"{task.project_id}/{task.id}/{task.name.split('.')[1]}/"
        object_storage_conf = object_storage_util.get_object_storage_config(self.storage_config_path)
        partitions = ''
        if hasattr(task, 'transform_partitions') and task.transform_partitions:
            partitions = ' PARTITION ( '
            values = task.transform_partitions[0]
            for index, value in enumerate(values):
                if index == 0:
                    partitions += value
                else:
                    partitions += f",{value}"
            partitions += ' ) '
        job_label = f"{task.name.split('.')[0]}_{task.name.split('.')[1]}_{datetime.now().strftime('%Y%m%d%H%M%S%f')}"
        unload_sql = f"EXPORT TABLE {task.name}\n" \
                     f"{partitions}\n " \
                     f"TO \"s3://{object_storage_conf['bucket']}/{file_path}\" \n" \
                     f"PROPERTIES(\n" \
                     f"\"label\" = \"{job_label}\" \n" \
                     f")\n" \
                     f"WITH S3\n" \
                     f"(\n" \
                     f"\"AWS_ENDPOINT\"=\"{object_storage_conf['endpoint']}\",\n" \
                     f"\"AWS_ACCESS_KEY\"=\"{object_storage_conf['id']}\",\n" \
                     f"\"AWS_SECRET_KEY\"=\"{object_storage_conf['key']}\",\n" \
                     f"\"AWS_REGION\"=\"{object_storage_conf['region']}\"\n" \
                     f");"
        logger.info(f"Unload data from Doris {self.connection_params['host']} with sql: {unload_sql}")
        try:
            self.execute_sql(unload_sql)
        except BaseException as e:
            logger.error(f"Unload data from Doris {self.connection_params['host']} failed, error: {e}")
            raise BaseException(f"Unload data from Doris {self.connection_params['host']} failed, error: {e}")
        check_unload_status_sql = f"SHOW EXPORT FROM {task.name.split('.')[0]} WHERE LABEL = '{job_label}';"
        while True:
            result = self.execute_sql(check_unload_status_sql)
            if len(result) > 0 and result[0]['State'] == 'FINISHED':
                break
            time.sleep(5)
        logger.info(f"Unload data from Doris {self.connection_params['host']} to path {file_path} successfully")
        return [file_path]

    def int_type_string(self):
        return 'INT'



import logging
import os
import random
from datetime import datetime
from time import sleep

import pymysql
from pymysqlpool.pool import Pool

from migration.connector.destination.base import Destination
from clickzetta.dbapi.connection import Connection as ClickZettaConnection
from clickzetta.client import Client
from migration.base.exceptions import DestinationExecutionError, GrammarRestrictionsError
from migration.base.status import Status
import migration.util.object_storage_util as object_storage_util

logger = logging.getLogger(__name__)

MAX_GET_STATUS_TIMES = 100


class DorisDestination(Destination):
    def __init__(self, config: dict, meta_conf_path=None, storage_conf_path=None):
        super().__init__('Doris', config)
        self.connection = None
        self.connection_params = self.get_connection_params()
        self.meta_conf_path = meta_conf_path
        self.storage_conf_path = storage_conf_path
        self.pool = Pool(host=self.connection_params['host'], port=self.connection_params['port'],
                         user=self.connection_params['user'], password=self.connection_params['password'],
                         max_size=50, autocommit=True)

    def get_connection_params(self):
        assert self.config['fe_servers']
        assert self.config['user']
        assert self.config['password']

        return {
            'host': self.config['fe_servers'][0].split(':')[0],
            'port': int(self.config['fe_servers'][0].split(':')[1]),
            'user': self.config['user'],
            'password': self.config['password']
        }

    def connect(self):
        if self.connection is None:
            self.connection = pymysql.connect(**self.connection_params)
            logger.info(f"Connect to Doris {self.connection_params['host']} successfully")

    def create_table(self, table_name, ddl: str):
        self.execute_sql(ddl)
        logger.info(f"Create table {table_name} successfully")

    def create_database(self, db_name):
        self.execute_sql(f"CREATE DATABASE IF NOT EXISTS {db_name}")
        logger.info(f"Create database {db_name} successfully")

    def execute_sql(self, sql, bind_params=None):
        try:
            connection = self.pool.get_conn()
            with connection.cursor() as cur:
                cur.execute(sql, bind_params)
                result = cur.fetchall()
                self.pool.release(connection)
                return result
        except Exception as e:
            logger.error("migration Error running SQL: {}, error:{}", sql, e)
            raise DestinationExecutionError('migration Error running SQL: {}, error:{}'.format(sql, e))

    def close(self):
        if self.connection is not None:
            self.connection.close()
        self.pool.destroy()

    def load_external_data(self, file_path, schema_name, table_name):
        object_storage = object_storage_util.get_object_storage_config(self.storage_conf_path)
        label_name = f"{schema_name}_{table_name}_{datetime.now().day}_{random.randint(1, 100000000)}"
        load_sql = f"use {schema_name};\n" \
                   f"LOAD LABEL {label_name}\n" \
                   f"(\n" \
                   f"DATA INFILE(\"s3://{object_storage['bucket']}/{file_path}/*\")\n" \
                   f"INTO TABLE {table_name}\n" \
                   f"FORMAT AS \"parquet\"\n" \
                   f")\n" \
                   f"WITH S3\n" \
                   f"(\n" \
                   f"\"AWS_ENDPOINT\" = \"{object_storage['endpoint']}\",\n" \
                   f"\"AWS_ACCESS_KEY\" = \"{object_storage['id']}\",\n" \
                   f"\"AWS_SECRET_KEY\" = \"{object_storage['key']}\",\n" \
                   f"\"AWS_REGION\" = \"{object_storage['region']}\"\n" \
                   f");\n"
        logger.info(f"Start to load data from {file_path} to {schema_name}.{table_name} with sql: {load_sql}")
        self.execute_sql(load_sql)
        current_times = 0
        while current_times < MAX_GET_STATUS_TIMES:
            if current_times == MAX_GET_STATUS_TIMES - 1:
                raise DestinationExecutionError(
                    f"doris load data from {file_path} to {schema_name}.{table_name}"
                    f" failed, get result more than 100 times.")
            check_state_sql = f"show load where Label = '{label_name}'"
            state = self.execute_sql(check_state_sql)[0]['State']
            if state == 'FINISHED':
                break
            elif state == 'CANCELLED' or state == 'FAILED':
                logger.error(f"doris load data from {file_path} to {schema_name}.{table_name} failed")
                raise DestinationExecutionError(
                    f"doris loading data from {file_path} to {schema_name}.{table_name} failed")
            else:
                logger.info(f"Waiting for doris S3 data loading, current state is {state}")
            sleep(5)
            current_times += 1

        logger.info(f"Doirs Load data from {file_path} to {schema_name}.{table_name} successfully")

    def gen_destination_ddl(self, db_name, table_name, columns, primary_keys, cluster_info, partition_columns):
        destination_ddl = 'CREATE TABLE IF NOT EXISTS {db}.{table} (\n'.format(db=db_name, table=table_name)
        for index, column in enumerate(columns):
            destination_ddl += '    {name} {type} {is_null} ' \
                .format(name=column.name,
                        type=column.type,
                        is_null='' if column.is_null else 'NOT NULL')
            # if column.default_value:
            #     destination_ddl += "   DEFAULT {default_value}".format(default_value=column.default_value)
            if index != len(columns) - 1:
                destination_ddl += ',\n'
            else:
                destination_ddl += '\n'

        if primary_keys:
            destination_ddl += '    UNIQUE KEY ({primary_keys}),\n'.format(primary_keys=','.join(primary_keys))
        destination_ddl += ') \n'
        if partition_columns:
            destination_ddl += 'PARTITIONED BY ({partitioned_by})\n'.format(partitioned_by=','.join(partition_columns))
        if cluster_info:
            destination_ddl += 'CLUSTERED BY ({clustered_by})\n'.format(
                clustered_by=','.join(cluster_info.cluster_keys))
            destination_ddl += 'SORTED BY ({clustered_by})\n'.format(clustered_by=','.join(cluster_info.cluster_keys))
            if cluster_info.bucket_num:
                destination_ddl += 'INTO {buckets} BUCKETS\n'.format(buckets=cluster_info.bucket_num)
            else:
                destination_ddl += 'INTO 32 BUCKETS\n'
        self.create_database(db_name)
        return destination_ddl



class ProfileConfigError(BaseException):
    """Raised when profile config is invalid."""


class SourceExecutionError(BaseException):
    """Raised when source execute error."""


class GrammarRestrictionsError(BaseException):
    """Raised when grammar error."""


class SchedulerRuntimeError(BaseException):
    """Raised when scheduler run error."""


class DestinationExecutionError(BaseException):
    """Raised when destination execute error."""


class TaskRuntimeError(BaseException):
    """Raised when task run error."""
